# copyright: (c) 2021 by Jesse Johnson.
# license: LGPL-3.0, see LICENSE.md for more details.
"""Manage version numbers."""

# import logging
from dataclasses import InitVar, asdict, dataclass, field
from typing import Any, Dict, List, Optional, Tuple, Union

from packaging.version import Version as PackageVersion
from packaging.version import _cmpkey, _parse_local_version, _Version
from transitions import Machine

# from transitions.extensions import HierarchicalMachine


@dataclass
class LocalConfig:
    """Manage local build versions."""

    states: List[Union[str, Dict[str, Any]]] = field(default_factory=list)
    transitions: List[Dict[str, Any]] = field(default_factory=list)

    def __post_init__(self) -> None:
        """Initialize local version configuration."""
        self.states = ['non_local', 'local']

        # add build number
        self.transitions.append(
            {
                'trigger': 'update_local',
                'source': 'non_local',
                'dest': 'local',
                'before': '_update_local',
            }
        )

        # update build number
        self.transitions.append(
            {
                'trigger': 'update_local',
                'source': 'local_local',
                'dest': '',
                'before': '_update_local',
            }
        )

        # remove build number
        self.transitions.append(
            {
                'trigger': 'update_local',
                'source': 'local',
                'dest': 'non_local',
                'before': '_update_local',
            }
        )


@dataclass
class ReleaseConfig:
    """Manage versioning flow."""

    default_release_type: InitVar[str]
    states: List[Union[str, Dict[str, Any]]] = field(default_factory=list)
    transitions: List[Dict[str, Any]] = field(default_factory=list)

    def __post_init__(self, default_release_type: str) -> None:
        """Initialize versioning config."""
        self.states = ['dev', 'alpha', 'beta', 'candidate', 'final', 'post']

        # XXX: need to calver as alternative first
        # self.transitions.append(
        #     {
        #         'trigger': 'bump_epoch',
        #         'source': '*',
        #         'dest': default_release_type,
        #         'before': '_bump_epoch',
        #         'after': 'new_release',
        #         'conditions': ['autostart_default_release'],
        #     }
        # )

        self.transitions.append(
            {
                'trigger': 'bump_major',
                'source': '*',
                'dest': default_release_type,
                'before': '_bump_major',
                'after': 'new_release',
                'conditions': ['autostart_default_release'],
            }
        )

        self.transitions.append(
            {
                'trigger': 'bump_minor',
                'source': '*',
                'dest': default_release_type,
                'before': '_bump_minor',
                'after': 'new_release',
                'conditions': ['autostart_default_release'],
            }
        )

        self.transitions.append(
            {
                'trigger': 'bump_micro',
                'source': '*',
                'dest': default_release_type,
                'before': '_bump_micro',
                'after': 'new_release',
                'conditions': ['autostart_default_release'],
            }
        )

        self.transitions.append(
            {
                'trigger': 'bump_epoch',
                'source': '*',
                'dest': 'final',
                'before': '_bump_epoch',
                # 'unless': ['autostart_default_release'],
            }
        )

        self.transitions.append(
            {
                'trigger': 'bump_major',
                'source': '*',
                'dest': 'final',
                'before': '_bump_major',
                'unless': ['autostart_default_release'],
            }
        )

        self.transitions.append(
            {
                'trigger': 'bump_minor',
                'source': '*',
                'dest': 'final',
                'before': '_bump_minor',
                'unless': ['autostart_default_release'],
            }
        )

        self.transitions.append(
            {
                'trigger': 'bump_micro',
                'source': '*',
                'dest': 'final',
                'before': '_bump_micro',
                'unless': ['autostart_default_release'],
            }
        )

        # development releases
        self.transitions.append(
            {
                'trigger': 'start_release',
                'source': ['final', 'post'],
                'dest': 'dev',
                'before': '_new_devrelease',
                'conditions': ['enable_devreleases'],
            }
        )

        # pre-releases
        self.transitions.append(
            {
                'trigger': 'start_release',
                'source': ['final', 'post'],
                'dest': 'alpha',
                'before': '_new_prerelease',
                'conditions': ['enable_prereleases'],
                'unless': ['enable_devreleases'],
            }
        )
        self.transitions.append(
            {
                'trigger': 'start_release',
                'source': 'dev',
                'dest': 'alpha',
                'before': '_new_prerelease',
                'conditions': ['enable_devreleases', 'enable_prereleases'],
            }
        )
        self.transitions.append(
            {
                'trigger': 'start_release',
                'source': 'alpha',
                'dest': 'beta',
                'before': '_new_prerelease',
                'conditions': ['enable_prereleases'],
            }
        )
        self.transitions.append(
            {
                'trigger': 'start_release',
                'source': 'beta',
                'dest': 'candidate',
                'before': '_new_prerelease',
                'conditions': ['enable_prereleases'],
            }
        )

        # final release
        self.transitions.append(
            {
                'trigger': 'start_release',
                'source': 'dev',
                'dest': 'final',
                'before': '_finalize_release',
                'conditions': ['enable_devreleases'],
                'unless': ['enable_prereleases'],
            }
        )

        self.transitions.append(
            {
                'trigger': 'start_release',
                'source': 'candidate',
                'dest': 'final',
                'before': '_finalize_release',
                'conditions': ['enable_prereleases'],
            }
        )
        self.transitions.append(
            {
                'trigger': 'start_release',
                'source': 'post',
                'dest': 'final',
                'before': '_finalize_release',
                'conditions': ['enable_postreleases'],
                'unless': ['enable_devreleases', 'enable_prereleases'],
            }
        )

        # post-releases
        self.transitions.append(
            {
                'trigger': 'start_postrelease',
                'source': 'final',
                'dest': 'post',
                'before': '_new_postrelease',
                'conditions': ['enable_postreleases'],
            }
        )

        # bump release number
        self.transitions.append(
            {
                'trigger': 'bump_release',
                'source': '*',
                'dest': None,
                'before': '_bump_release',
            }
        )


class Version(PackageVersion):
    """Provide PEP440 compliant versioning."""

    def __init__(self, version: str, **kwargs: Any) -> None:
        """Initialize version object."""
        self.compat = kwargs.pop('compat', 'pep440')
        super().__init__(version=version)

        # TODO: transitions here should be populated by VCS workflow

        self.enable_devreleases = kwargs.get('enable_devreleases', True)
        self.enable_prereleases = kwargs.get('enable_prereleases', True)
        self.enable_postreleases = kwargs.get('enable_postreleases', True)
        self.autostart_default_release = kwargs.get(
            'autostart_default_release', True
        )

        _release = ReleaseConfig(
            default_release_type=self.default_release_type,
        )
        self.machine = Machine(
            self,
            **asdict(_release),
            initial=self.release_type,
        )
        self.local_machine = Machine(
            self,
            **asdict(LocalConfig()),
            initial='local' if self.local else 'non_local',
            model_attribute='build',
        )

        self.machine.auto_transitions = False

    def __str__(self) -> str:
        """Handle version formatting."""
        parts = []

        # Epoch
        if self.epoch != 0:
            parts.append(f"{self.epoch}!")

        # Release segment
        parts.append('.'.join(str(x) for x in self.release))

        # Pre-release
        if self.pre is not None:
            if self.compat == 'semver':
                separator = '-'
            else:
                separator = ''
            parts.append(f"{separator}{self.pre[0]}{str(self.pre[1])}")

        # Post-release
        if self.post is not None:
            if self.compat == 'semver':
                separator = '-'
            else:
                separator = '.post'
            parts.append(f"{separator}{self.post}")

        # Development release
        if self.dev is not None:
            if self.compat == 'semver':
                separator = '-'
            else:
                separator = '.'
            parts.append(f"{separator}dev{self.dev}")

        # Local version segment
        if self.local is not None:
            parts.append(f"+{self.local}")

        return ''.join(parts)

    @property
    def release_type(self) -> str:
        """Get the current state of package release."""
        if self.is_devrelease:
            state = 'dev'
        elif self.is_prerelease and self.pre:
            if self.pre[0] == 'a':
                return 'alpha'
            if self.pre[0] == 'b':
                return 'beta'
            if self.pre[0] == 'rc':
                return 'candidate'
        elif self.is_postrelease:
            state = 'post'
        else:
            state = 'final'
        return state

    @property
    def default_release_type(self) -> str:
        """Get the starting release type."""
        if self.enable_devreleases:
            return 'dev'
        if self.enable_prereleases:
            return 'alpha'
        return 'final'

    @property
    def query(self) -> str:
        """Get PEP-440 compliant regex query for version."""
        r = '.'.join(str(x) for x in self.release)
        if self.dev:
            v = f"{r}[-_\\.]?dev[-_\\.]?{self.dev or '0?'}"
            return v

        if self.pre:
            if self.epoch > 0:
                v = f"{self.epoch}!{r}"
            (pre, inst) = self.pre
            if pre == 'a':
                v = f"{r}[-_\\.]?(?:a|alpha)[-_\\.]?{inst or '0?'}"
            if pre == 'b':
                v = f"{r}[-_\\.]?(?:b|beta)[-_\\.]?{inst or '0?'}"
            if pre == 'rc':
                candidate = 'c|candidate|rc|release'
                v = f"{r}[-_\\.]?(?:{candidate})[-_\\.]?{inst or '0?'}"
            return v
        if self.post:
            v = f"{r}[-_\\.]?(?:post[-_\\.]?)?{self.post}"
            return v
        return str(self)

    def __update_version(
        self,
        epoch: Optional[int] = None,
        release: Optional[Tuple[Any, ...]] = None,
        pre: Optional[Tuple[str, int]] = None,
        post: Optional[Tuple[str, int]] = None,
        dev: Optional[Tuple[str, int]] = None,
        local: Optional[str] = None,
    ) -> None:
        """Update the internal version state."""
        if not (epoch or release):
            pre = pre or self.pre
            post = post or (None if self.post is None else ('post', self.post))
            dev = dev or (None if self.dev is None else ('dev', self.dev))

        self._version = _Version(
            epoch=epoch or self.epoch,
            release=release or self.release,
            pre=pre,
            post=post,
            dev=dev,
            local=_parse_local_version(local) if local else None,
        )

        self._key = _cmpkey(
            self._version.epoch,
            self._version.release,
            self._version.pre,
            self._version.post,
            self._version.dev,
            self._version.local,
        )

    def _bump_epoch(self) -> None:
        """Update epoch releaes for version system changes."""
        self.__update_version(epoch=self.epoch + 1)

    def _bump_major(self) -> None:
        """Update major release to next version number."""
        self.__update_version(release=(self.major + 1, 0, 0))

    def _bump_minor(self) -> None:
        """Update minor release to next version number."""
        self.__update_version(release=(self.major, self.minor + 1, 0))

    def _bump_micro(self) -> None:
        """Update micro release to next version number."""
        self.__update_version(release=(self.major, self.minor, self.micro + 1))

    def __bump_version(self, segment: str) -> None:
        """Bump version based on version kind."""
        if segment == 'epoch':
            self._bump_epoch()
        if segment == 'major':
            self._bump_major()
        if segment == 'minor':
            self._bump_minor()
        if segment == 'micro':
            self._bump_micro()

    def _new_devrelease(self, segment: Optional[str] = None) -> None:
        """Update to the next development release version number."""
        if self.dev is None:
            if segment:
                self.__bump_version(segment)
            self.__update_version(dev=('dev', 0))

    def _new_prerelease(self, segment: Optional[str] = None) -> None:
        """Update to next prerelease version type."""
        if self.pre is not None:
            if self.pre[0] == 'a':
                pre = ('b', 0)
            elif self.pre[0] == 'b':
                pre = ('rc', 0)
        else:
            if segment:
                self.__bump_version(segment)
            pre = ('a', 0)
        self.__update_version(pre=pre)

    def _new_postrelease(self) -> None:
        """Update the post release version number."""
        if self.release_type == 'final':
            self.__update_version(post=('post', 0))

    def new_release(
        self,
        segment: Optional[str] = None,
        kind: Optional[str] = None,
    ) -> None:
        """Update the version release."""
        kind = kind or self.default_release_type
        if self.enable_devreleases and kind == 'dev':
            self._new_devrelease(segment=segment)
        if kind == 'alpha':
            self._new_prerelease(segment=segment)
        if kind == 'final':
            self._finalize_release()

    def _bump_release(self, *args: Any, **kwargs: Any) -> None:
        """Update to the next development release version number."""
        if self.enable_devreleases and self.dev is not None:
            dev = self.dev + 1
            self.__update_version(dev=('dev', dev))
        elif self.enable_prereleases and self.pre is not None:
            pre = (self.pre[0], self.pre[1] + 1)
            self.__update_version(pre=pre)
        elif self.enable_postreleases:
            if self.release_type == 'final':
                self.__update_version(post=('post', 0))
            elif self.post is not None:
                post = self.post + 1
                self.__update_version(post=('post', post))

    def _new_local(self, local: str) -> None:
        """Create new local version instance number."""
        self.__update_version(local=local)

    def _finalize_release(self, segment: Optional[str] = None) -> None:
        """Update to next prerelease version type."""
        self.__update_version(release=(self.major, self.minor, self.micro))



# type: ignore
"""Test versioning capability."""

from versioning.version import Version


def test_versioning():
    """Test capability to bump versions."""
    v = Version('1.0.0', autostart_default_release=False)
    assert v.state == 'final'
    assert v.major == 1
    assert v.minor == 0
    assert v.micro == 0
    v.bump_major()
    assert str(v) == '2.0.0'
    v.bump_minor()
    assert str(v) == '2.1.0'
    v.bump_micro()
    assert str(v) == '2.1.1'


def test_devrelease():
    """Test development releases."""
    v = Version('1.0.0', enable_prereleases=False)
    assert v.default_release_type == 'dev'
    assert v.enable_prereleases is False
    v.start_release(segment='minor')
    assert v == Version('1.1.0.dev0')
    v.bump_release()
    assert str(v) == '1.1.0.dev1'
    v.start_release()
    assert str(v) == '1.1.0'


def test_devrelease_state():
    """Test development release state."""
    v = Version('1.0.0', enable_prereleases=False)
    assert v.default_release_type == 'dev'
    assert v.enable_prereleases is False
    assert v.state == 'final'
    v.start_release()
    assert v.state == 'dev'
    v.start_release()
    assert v.state == 'final'


def test_prerelease():
    """Test pre-releases."""
    v = Version('2.0.0', enable_devreleases=False)
    assert v.default_release_type == 'alpha'
    assert v.state == 'final'
    v.start_release(segment='major')
    assert str(v) == '3.0.0a0'
    assert v.pre == ('a', 0)
    v.bump_release()
    assert str(v) == '3.0.0a1'
    v.start_release()
    assert str(v) == '3.0.0b0'
    v.bump_release()
    assert str(v) == '3.0.0b1'
    v.start_release()
    assert str(v) == '3.0.0rc0'
    v.bump_release()
    assert str(v) == '3.0.0rc1'


def test_prerelease_states():
    """Test pre-release states."""
    v = Version('1.0.0', enable_devreleases=False)
    assert v.default_release_type == 'alpha'
    assert v.state == 'final'
    v.start_release()
    assert v.state == 'alpha'
    v.start_release()
    assert v.state == 'beta'
    v.start_release()
    assert v.state == 'candidate'
    v.start_release()
    assert v.state == 'final'


def test_post():
    """Test post-release version bump."""
    v = Version('2.0.0', autostart_default_release=True)
    assert v.state == 'final'
    v.start_postrelease()
    assert v.release == (2, 0, 0)
    assert v.base_version == '2.0.0'
    assert v.major == 2
    assert v.minor == 0
    assert v.micro == 0
    assert v.state == 'post'
    v.bump_release()
    assert v.state == 'post'
    v.bump_major()
    assert v.state == 'dev'


def test_post_state():
    """Test post-release state."""
    v = Version('2.0.0')
    assert v.state == 'final'

    v.start_postrelease()
    assert v.state == 'post'


def test_local():
    """Test local version update."""
    v = Version('3.0.0+1234567', autostart_default_release=False)
    assert v.release == (3, 0, 0)
    assert v.pre is None
    assert v.post is None
    assert v.local == '1234567'
    v.bump_major()
    assert str(v) == '4.0.0'


def test_local_state():
    """Test local version state."""
    v = Version('1.0.0')
    assert v.state == 'final'

    v1 = Version(f"{str(v)}+build.0")
    assert v1.local == 'build.0'
    assert v1.state == 'final'

    v._new_local('4321')
    assert v == Version('1.0.0+4321')
    assert v.local == '4321'
    assert v.state == 'final'

    # TODO: assert build is gone after bump


def test_local_removed():
    """Test removal of local version."""
    v = Version('1.0.0+build.0')
    assert v.local == 'build.0'
    v.state == 'final'

    # ensure local version is removed from development
    v.start_release()
    v = Version('1.0.0.dev')
    assert v.state == 'dev'
    assert v.local is None

    # ensure local version is removed from final
    v = Version('1.0.0+build.0', autostart_default_release=True)
    v.bump_minor()
    assert v.state == 'dev'
    assert v.local is None


def test_mixed():
    """Test mixed versions."""
    # TODO: getting warning due to both
    v = Version('5.0.0-rc5+build254')
    assert v.state == 'candidate'
    assert v.local == 'build254'


def test_epoch():
    """Test epoch version bump."""
    v = Version('1!6.0.0')
    assert v.state == 'final'
    assert v.epoch == 1
    v.bump_epoch()
    assert str(v) == '2!6.0.0'



# copyright: (c) 2021 by Jesse Johnson.
# license: LGPL-3.0, see LICENSE.md for more details.
"""Parse git commit messages."""

# import logging
import os
from typing import Any, List, Optional

from pygit2 import (
    GIT_OBJECT_COMMIT,
    Commit,
    RemoteCallbacks,
    Repository,
    Signature,
    Tag,
    UserPass,
)

from versioning.exception import VersioningException


class Git:
    """Provide settings for git repositories."""

    def __init__(self, repo: Repository) -> None:
        """Initialize git object."""
        self.repo = repo
        self.hooks_dir = os.path.join(self.repo.path, 'hooks')
        self.config = os.path.join(self.repo.path, 'config')

    @property
    def username(self) -> str:
        """Return username from git configuration."""
        try:
            return self.repo.config['user.name']
        except Exception as exc:
            raise VersioningException('git user.name not configured') from exc

    @property
    def branch(self) -> str:
        """Return branch name of current branch."""
        return self.repo.head.shorthand

    @property
    def email(self) -> str:
        """Return email from git configuration."""
        try:
            return self.repo.config['user.email']
        except Exception as exc:
            raise VersioningException('git user.email not configured') from exc

    @property
    def repo_dir(self) -> str:
        """Return directory of repository."""
        return self.repo.path

    @property
    def working_dir(self) -> str:
        """Return working directory of project."""
        return os.path.abspath(os.path.join(self.repo_dir, os.pardir))

    @property
    def ref(self) -> str:
        """Retrieve the current reference."""
        return self.repo.head.name

    def add_remote(
        self,
        name: str,
        url: str,
        fetch: Optional[str] = None,
        # **kwargs: Any,
    ) -> None:
        """Add remote."""
        if name not in [x.name for x in self.repo.remotes]:
            if fetch is None:
                fetch = f"+refs/heads/*:refs/remotes/{name}/*"
            self.repo.remotes.create(name, url, fetch)
        elif self.repo.remotes[name].url != url:
            self.repo.remotes.set_url(name, url)

    def commit(
        self,
        branch: Optional[str] = None,
        filepaths: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> Optional[Commit]:
        """Create commit."""
        ref = f"refs/heads/{branch}" if branch else self.ref

        author = Signature(self.username, self.email)
        committer = kwargs.get('commiter', author)
        message = kwargs.get('message', 'ci: bump version')

        # populate index
        index = self.repo.index
        if not filepaths:
            index.add_all()
        else:
            for filepath in filepaths:
                index.add(
                    os.path.relpath(
                        filepath, os.path.join(self.repo.path, os.pardir)
                    )
                )
        index.write()
        tree = index.write_tree()

        # get parent
        parents = kwargs.get('parents', [self.repo.head.target])
        encoding = kwargs.get('encoding', 'utf-8')

        # commit
        if not kwargs.get('dry_run', False):
            commit = self.repo.create_commit(
                ref,
                author,
                committer,
                message,
                tree,
                parents,
                encoding,
            )
            return commit
        # TODO: should a mock commit be created?
        return None

    def tag(
        self, name: str, branch: Optional[str] = None, **kwargs: Any
    ) -> Optional[Tag]:
        """Create tag."""
        ref = f"refs/heads/{branch}" if branch else self.ref
        commit = self.repo.resolve_refish(ref)[0]
        oid = str(commit.id)
        kind = kwargs.get('kind', GIT_OBJECT_COMMIT)

        # tagger = pygit2.Signature('Alice Doe', 'adoe@example.com', 12347, 0)
        tagger = Signature(self.username, self.email)
        message = kwargs.get('message')
        if not kwargs.get('dry_run', False):
            tag = self.repo.create_tag(
                name, oid, kind, tagger, message or f"ci: {name}"
            )
            return tag
        # TODO: should a mock tag be created?
        return None

    def push(
        self,
        branch: Optional[str] = None,
        remote: str = 'origin',
        remote_branch: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> None:
        """Push commited changes."""
        # TODO: feels like remote should be encapsulated with creds
        remote_branch = remote_branch or branch or self.ref.split('/')[-1]
        # ref = f"refs/heads/{branch}" if branch else self.ref
        # remote_ref = f"refs/remotes/{remote}/{remote_branch}"
        # print(ref, remote_ref)

        remote_repo = self.repo.remotes[remote]
        credentials = UserPass(username or self.username, password)
        remote_repo.credentials = credentials

        callbacks = RemoteCallbacks(credentials=credentials)
        remote_repo.push([f"refs/heads/{branch}"], callbacks=callbacks)



# copyright: (c) 2021 by Jesse Johnson.
# license: LGPL-3.0, see LICENSE.md for more details.
"""Initialize versioning instances."""

import logging
from typing import Any

from pygit2 import Repository

from versioning.config import CONFIG_FILES, REPO_DIR, Config
from versioning.controller import ReleaseController
from versioning.exception import VersioningException
from versioning.vcs import Git
from versioning.version import Version  # noqa

__author__ = 'Jesse P. Johnson'
__author_email__ = 'jpj6652@gmail.com'
__title__ = 'proman-versioning'
__description__ = 'Workflows to manage project versioning.'
__version__ = '0.6.3a0'
__license__ = 'LGPL-3.0'
__copyright__ = 'Copyright 2021 Jesse Johnson.'
__all__ = (
    'ReleaseController',
    'get_release_controller',
    'Version',
)

logging.getLogger(__name__).addHandler(logging.NullHandler())


def get_release_controller(**kwargs: Any) -> ReleaseController:
    """Create and return a release controller."""
    try:
        repo_dir = kwargs.pop('repo_dir', REPO_DIR)
        repo = Git(Repository(repo_dir))

        config_files = kwargs.pop('config_files', CONFIG_FILES)
        config = Config(filepaths=config_files)

        return ReleaseController(config=config, repo=repo, **kwargs)
    except Exception as err:
        raise VersioningException(err) from err



# copyright: (c) 2021 by Jesse Johnson.
# license: LGPL-3.0, see LICENSE.md for more details.
"""Control project versions."""

import logging
import sys
from typing import Optional

from versioning import get_release_controller

_log = logging.getLogger(__name__)
_controller = get_release_controller()


def bump(
    commit: bool = False,
    message: Optional[str] = None,
    tag: bool = False,
    tag_name: Optional[str] = None,
    # sign: bool = False,
    release: bool = False,
    build: Optional[str] = None,
    dry_run: bool = False,
) -> None:
    """Update project version.

    Parameters
    ----------
    build: str
        Build version from other systems the may help coordinate build
        artifacts.
    commit: bool
        Commit release changes to project.
    dry_run: bool
        Run the command without making changes.
    message: str
        Notes included with the commit.
    release: bool
        Make this version a new release.
    tag: bool
        Tag this version.
    tag_name: str
        The tag name to use for this tag.

    """
    # sign: bool
    #     Sign commit with PKI signature.
    version = _controller.update_version(
        commit=commit,
        release=release,
        tag=tag,
        tag_name=tag_name,
        # sign_tag=sign,
        message=message,
        build=build,
        dry_run=dry_run,
    )
    print(str(version), file=sys.stdout)


def info(
    release: bool = False,
    filepaths: bool = False,
    compat: Optional[str] = None,
) -> None:
    """Get current version project.

    Parameters
    ----------
    release: bool
        View the current release type
    filepaths: bool
        List the version filepaths
    compat: str:
        Type of version format

    """
    if release:
        print(_controller.release, file=sys.stdout)
    if filepaths:
        for x in _controller.config.templates:
            print(x['filepath'], file=sys.stdout)
    if compat:
        ...
    if not release and not filepaths:
        print(_controller.config.version, file=sys.stdout)


def push(
    branch: Optional[str] = None,
    remote: str = 'origin',
    remote_branch: Optional[str] = None,
    remote_url: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
) -> None:
    """Push project version.

    Parameters
    ----------
    branch: str
        The name of the local branch to push.
    remote: str
        The reference name of the repository.
    remote_branch: str
        The branch name of the remote reference.
    remote_url: str
        Remote URL of the repository.
    username: str
        Username for access to remote repository.
    password: str
        Password for access to remote repository.

    """
    # INFO: Helper for when run from containers without git.
    _controller.push_changes(
        branch=branch,
        remote=remote,
        remote_branch=remote_branch,
        remote_url=remote_url,
        username=username,
        password=password,
    )



# type: ignore
"""Test git hooks pipeline."""

import os
from unittest.mock import Mock, mock_open, patch

from versioning import ReleaseController, Version
from versioning.config import Config
from versioning.vcs import Git

config = Config(
    filepaths=[],
    defaults={
        # 'project': {'version': '1.2.3'},
        'proman': {
            'version': '1.2.3',
            'versioning': {
                'files': [
                    {
                        'filepath': 'pyproject.toml',
                        'pattern': 'version = "${version}"',
                    },
                    {
                        'filepath': 'chart/Chart.yaml',
                        'compat': 'semver',
                        'patterns': [
                            'version: "${version}"',
                            'appVersion: "${version}"',
                        ],
                    },
                ],
            },
        }
    },
)


@patch('builtins.open', new_callable=mock_open, read_data='version = "1.2.3"')
def test_create_devrelease(mock_file):
    """Test development release creation."""
    working_dir = os.path.join(os.sep, 'mock', '.git')
    controller = ReleaseController(
        repo=Git(Mock(path=working_dir, head={'name': 'mock'})),
        config=config,
        message='release: create devrelease',
    )
    assert controller.config.version == Version('1.2.3')

    controller.update_version(
        commit=False,
        # release=release,
        # tag=tag,
        # tag_name=tag_name,
        # message=message,
        # sign_tag=sign,
        # build=build,
        dry_run=True,
    )
    assert controller.config.version == Version('1.3.0.dev0')
